// Copyright (C) 2018  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package main
// misc utilities

import (
	"context"

	"github.com/hanwen/go-fuse/fuse"
	"github.com/hanwen/go-fuse/fuse/nodefs"

	"lab.nexedi.com/kirr/neo/go/zodb"
)

// asctx represents fuse context as context.Context ready for interrupt handling.
//
// XXX temp. only after proper interrupt handling is not yet merged into go-fuse.
//     https://github.com/hanwen/go-fuse/pull/15
func asctx(fctx *fuse.Context) context.Context {
	// FIXME stub
	return context.Background()
}


// NewStaticFile creates nodefs.Node for file with static data.
func NewStaticFile(data []byte) *SmallFile {
	return NewSmallFile(func() []byte {
		return data
	})
}

// SmallFile is a nodefs.Node for file with dynamic, but always small, data.
type SmallFile struct {
	nodefs.Node

	// readData gives whole file data
	readData func() []byte
}

func NewSmallFile(readData func() []byte) *SmallFile {
	return &SmallFile{Node: nodefs.NewDefaultNode(), readData: readData}
}

func (f *SmallFile) GetAttr(out *fuse.Attr, _ nodefs.File, _ *fuse.Context) fuse.Status {
	data := f.readData()
	out.Size = uint64(len(data))
	out.Mode = fuse.S_IFREG | 0644
	return fuse.OK
}

func (f *SmallFile) Read(_ nodefs.File, dest []byte, off int64, _ *fuse.Context) (fuse.ReadResult, fuse.Status) {
	data := f.readData()
	l := int64(len(data))
	end := off + l
	if end > l {
		end = l
	}

	return fuse.ReadResultData(data[off:end]), fuse.OK
}


// mkdir adds child to parent as directory.
//
// Note: parent must be already in the filesystem tree - i.e. associated
// with inode. if not - nodefs will panic in Inode.NewChild on nil dereference.
func mkdir(parent nodefs.Node, name string, child nodefs.Node) {
	parent.Inode().NewChild(name, true, child)
}

// mkfile adds child to parent as file.
//
// Note: parent must be already in the filesystem tree (see mkdir for details).
func mkfile(parent nodefs.Node, name string, child nodefs.Node) {
	parent.Inode().NewChild(name, false, child)
}


// mount is like nodefs.MountRoot but allows to pass in full fuse.MountOptions.
func mount(mntpt string, root nodefs.Node, opts *fuse.MountOptions) (*fuse.Server, *nodefs.FileSystemConnector, error) {
	nodefsOpts := nodefs.NewOptions()
	nodefsOpts.Debug = opts.Debug

	fsconn := nodefs.NewFileSystemConnector(root, nodefsOpts)

	fssrv, err := fuse.NewServer(fsconn.RawFS(), mntpt, opts)
	if err != nil {
		return nil, nil, err
	}

	return fssrv, fsconn, nil
}


// typeOf returns ZODB type of an object.
//
// it differs from %T in fmt in that it supports printing zodb.Broken with
// details properly.
func typeOf(obj interface{}) string {
	return zodb.ClassOf(obj)
}
