# -*- coding: utf-8 -*-
# Copyright (C) 2018  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""test wcfs filesystem from outside as python client process"""

from wendelin.lib.testing import getTestDB
from wendelin.lib.zodb import dbclose
from wendelin.lib.mem import memcpy
from wendelin.bigfile.file_zodb import ZBigFile
from wendelin.bigfile.tests.test_filezodb import blksize
from wendelin import wcfs

import transaction
from persistent import Persistent
from persistent.timestamp import TimeStamp

import os, os.path, sys, subprocess
from errno import EINVAL
from golang import func, method, defer
from pytest import raises

testdb = None
testzurl = None     # URL of testdb
testmntpt = None    # wcfs is mounted here
def setup_module():
    global testdb, testzurl, testmntpt
    testdb = getTestDB()
    testdb.setup()

    zstor = testdb.getZODBStorage()
    testzurl = wcfs._zstor_2zurl(zstor)
    zstor.close()
    testmntpt = wcfs._mntpt_4zurl(testzurl)
    os.rmdir(testmntpt)

def teardown_module():
    testdb.teardown()

# make sure we start every test without wcfs server running.
def setup_function(f):
    assert not os.path.exists(testmntpt)

# make sure we unmount wcfs after every test.
def teardown_function(f):
    mounted = not subprocess.call(["mountpoint", "-q", testmntpt])
    if mounted:
        subprocess.check_call(["fusermount", "-u", testmntpt])
    if os.path.exists(testmntpt):
        os.rmdir(testmntpt)


# readfile reads file @ path.
def readfile(path):
    with open(path) as f:
        return f.read()

# tidtime converts tid to transaction commit time.
def tidtime(tid):
    t = TimeStamp(tid).timeTime()

    # ZODB/py vs ZODB/go time resolution is not better than 1µs
    # see e.g. https://lab.nexedi.com/kirr/neo/commit/9112f21e
    #
    # NOTE pytest.approx supports only ==, not e.g. <, so we use plain round.
    return round(t, 6)


# check that zurl does not change from one open to another storage open.
def test_zurlstable():
    for i in range(10):
        zstor = testdb.getZODBStorage()
        zurl  = wcfs._zstor_2zurl(zstor)
        zstor.close()
        assert zurl == testzurl

@func
def test_join():
    zurl = testzurl
    with raises(RuntimeError, match="wcfs: join .*: server not started"):
        wcfs.join(zurl, autostart=False)

    wc = wcfs._start(zurl)
    defer(wc.close)
    assert wc.mountpoint == testmntpt
    assert readfile(wc.mountpoint + "/.wcfs") == zurl
    assert os.path.isdir(wc.mountpoint + "/bigfile")

    wc2 = wcfs.join(zurl, autostart=False)
    defer(wc2.close)
    assert wc2.mountpoint == wc.mountpoint

@func
def test_join_autostart():
    zurl = testzurl
    with raises(RuntimeError, match="wcfs: join .*: server not started"):
        wcfs.join(zurl, autostart=False)

    wc = wcfs.join(zurl, autostart=True)
    defer(wc.close)
    assert wc.mountpoint == testmntpt
    assert readfile(wc.mountpoint + "/.wcfs") == zurl
    assert os.path.isdir(wc.mountpoint + "/bigfile")


# XXX hack (tmp ?)
# Conn._sync makes sure that underlying wcfs is synced to last ZODB data
@method(wcfs.Conn)
def _sync(self):
    print >>sys.stderr, "\n# >>> wcfs.Conn.sync"
    zurl = readfile(self.mountpoint + "/.wcfs")
    bigfilev = os.listdir(self.mountpoint + "/bigfile")
    self.close()

    subprocess.check_call(["fusermount", "-u", self.mountpoint])
    c2 = wcfs.join(zurl, autostart=True)
    assert c2.mountpoint == self.mountpoint
    self._fwcfs = c2._fwcfs

    # recreate bigfile entries
    for bf in bigfilev:
        os.mkdir("%s/bigfile/%s" % (self.mountpoint, bf))

    print >>sys.stderr, "# <<< wcfs.Conn.sync\n"


# XXX parametrize zblk0, zblk1
# XXX select !wcfs mode so that we prepare data through !wcfs path.
@func
def test_bigfile_empty():
    root = testdb.dbopen()
    defer(lambda: dbclose(root))

    root['zfile'] = f = ZBigFile(blksize)

    # NOTE there is no clean way to retrieve tid of just committed transaction
    #      we'll be using last._p_serial as workaround.
    root['last'] = last = Persistent()
    transaction.commit()
    tid1 = last._p_serial

    last._p_changed = 1
    transaction.commit()
    tid2 = last._p_serial

    assert tidtime(tid2) > tidtime(tid1)

    wc = wcfs.join(testzurl, autostart=True)
    defer(wc.close)

    # path to bigfile/ under wcfs
    bigpath = wc.mountpoint + "/bigfile"

    # mkdir to non-BigFile - must be rejected
    with raises(OSError) as exc:
        os.mkdir("%s/%s" % (bigpath, last._p_oid.encode('hex')))
    assert exc.value.errno == EINVAL


    # path to f under wcfs
    fpath = "%s/%s" % (bigpath, f._p_oid.encode('hex'))

    os.mkdir(fpath)
    st = os.stat(fpath + "/head/data")
    assert st.st_size == 0
    assert st.st_mtime == tidtime(tid1)

    # head/at = last txn of whole db
    assert readfile(fpath + "/head/at") == tid2.encode('hex')

    # TODO check head/at syncs to later non-bigfile commits


    # commit data to f and make sure we can see it on wcfs
    #hole = 10      XXX reenable
    hole = 1
    fh  = f.fileh_open()
    vma = fh.mmap(hole, 1)    # 1 page at offset=10
    s   = b"hello world"
    memcpy(vma, s)

    last._p_changed = 1
    transaction.commit()
    tidlast = last._p_serial

    # sync wcfs to ZODB
    wc._sync()

    # we wrote "hello world" after 10th block, but size is always mutiple of blksize.
    fsize = (hole + 1)*blksize

    st = os.stat(fpath + "/head/data")
    assert st.st_size  == fsize
    #assert st.st_mtime == tidtime(tidlast)     FIXME proper sync
    assert readfile(fpath + "/head/at") == tidlast.encode("hex")

    data = readfile(fpath + "/head/data")
    assert len(data) == fsize
    for i in range(hole):
        assert data[i*blksize:(i+1)*blksize] == b'\0'*blksize
    tail = data[hole*blksize:]
    assert tail[:len(s)] == s
    assert tail[len(s):] == b'\0'*(blksize - len(s))


    # commit data again and make sure we can see both latest and snapshotted states.
    tcommit1 = tidlast

    fh  = f.fileh_open()
    vma1 = fh.mmap(hole, 1)
    vma2 = fh.mmap(hole+1, 1)
    s1  = b"hello 123"
    s2  = b"alpha"
    memcpy(vma1,s1)
    memcpy(vma2,s2)

    last._p_changed = 1
    transaction.commit()
    tidlast = last._p_serial

    wc._sync()

    fsize1 = fsize
    fsize  = fsize1 + blksize   # we added one more block

    st = os.stat(fpath + "/head/data")
    assert st.st_size  == fsize
    #assert st.st_mtime == tidtime(tidlast)     FIXME proper sync
    assert readfile(fpath + "/head/at") == tidlast.encode("hex")

    data = readfile(fpath + "/head/data")
    assert len(data) == fsize
    for i in range(hole):
        assert data[i*blksize:(i+1)*blksize] == b'\0'*blksize
    tail1 = data[hole*blksize:(hole+1)*blksize]
    assert tail1[:len(s1)] == s1
    assert tail1[len(s1):len(s)] == "ld"
    assert tail1[len(s):] == b'\0'*(blksize - len(s))
    tail2 = data[(hole+1)*blksize:]
    assert tail2[:len(s2)] == s2
    assert tail2[len(s2):] == b'\0'*(blksize - len(s2))

    return  # XXX temp

    # path to f's state @tcommit1
    fpath1 = fpath + ("/@%s" % tcommit1.encode("hex"))

    os.mkdir(fpath1)
    st = os.stat(fpath1 + "/data")
    assert st.st_size  == fsize1
    #assert st.st_mtime == tidtime(tcommit1)     FIXME proper sync
    #assert readfile(fpath + "/at") == tcommit1.encode("hex")   XXX do we need it?

    data = readfile(fpath1 + "/data")
    assert len(data) == fsize1
    for i in range(hole):
        assert data[i*blksize:(i+1)*blksize] == b'\0'*blksize
    tail = data[hole*blksize:]
    assert tail[:len(s)] == s
    assert tail[len(s):] == b'\0'*(blksize - len(s))


# TODO pagecache state after loading (via mincore)
